# coding: utf-8

from __future__ import unicode_literals

import os
import re
import logging

from tionix_autosetup.executors.base import Executor
from tionix_autosetup.executors.config import DropConfig
from tionix_autosetup.helpers import Platform

logger = logging.getLogger(__name__)


class ApacheRestart(Executor):

    apache_commands = {
        Platform.UBUNTU: ('service apache2 restart',),
        Platform.ALT_LINUX: ('systemctl restart httpd2',),
        Platform.CENTOS: ('systemctl restart httpd',),
    }

    @property
    def commands(self):
        return self.apache_commands.get(Platform.distribution)


class DropApacheConfig(DropConfig):

    config_dirs = {
        Platform.UBUNTU: ('/etc/apache2/conf-enabled',),
        Platform.ALT_LINUX: ('/etc/httpd2/conf/sites-enabled',),
        Platform.CENTOS: ('/etc/httpd/conf.d',),
    }

    force_config_dirs = {
        Platform.UBUNTU: ('/etc/apache2/conf-available',),
        Platform.ALT_LINUX: ('/etc/httpd2/conf/sites-available',),
        Platform.CENTOS: ('/etc/httpd/conf.d',),
    }

    config_name = None

    @property
    def paths(self):
        config_paths = [
            os.path.join(config_dir, self.config_name)
            for config_dir in self.config_dirs.get(Platform.distribution)]
        if self.force:
            config_paths.extend([
                os.path.join(force_config_dir, self.config_name)
                for force_config_dir
                in self.force_config_dirs.get(Platform.distribution)])
        return config_paths

    def pre_check(self, *arg, **kwargs):
        return True


class UpdateApacheConfig(Executor):
    config_path = None
    user_pattern = re.compile(r'user=[a-z0-9_-]+\s+group=[a-z0-9_-]+\s+')

    @property
    def procesors(self):
        """Configuration file processors list.

         To extend processors list just define method starting with "update_"
        """
        data_processors = []
        for attr_name in dir(self):
            if attr_name.startswith('update_'):
                attr = getattr(self, attr_name)
                if callable(attr):
                    data_processors.append(attr)
        return data_processors

    def update_user(self, data):
        replace_str = 'user={0} group={0} '.format(
            self.variables_dict.get('TNX_USERNAME'))
        if replace_str not in data:
            data = re.sub(self.user_pattern, replace_str, data)
        return data

    def _run(self, **kwargs):
        if os.path.exists(self.config_path):
            with open(self.config_path) as fh:
                data = fh.read()
            for processor in self.procesors:
                data = processor(data)
            with open(self.config_path, 'w') as fh:
                fh.write(data)
        else:
            logger.error("File {0} doesn't exist".format(self.config_path))
        return


class ApacheConfigPathMixin(object):
    config_name = None

    paths = {
        Platform.UBUNTU: ('/etc/apache2/conf-available',),
        Platform.ALT_LINUX: ('/etc/httpd2/conf/sites-available',),
        Platform.CENTOS: ('/etc/httpd/conf.d',),
    }

    @property
    def config_path(self):

        path = next((i for i in self.paths.get(Platform.distribution)
                     if os.path.exists(i)), None)
        if path:
            path = os.path.join(path, self.config_name)
        return path


class CopyApacheConfig(Executor):

    cfg_name = ''

    @property
    def commands(self):
        copy_commands = {
            Platform.UBUNTU: (
                'ln -s /etc/apache2/conf-available/{0} '
                '/etc/apache2/conf-enabled/{0}'.format(self.cfg_name),),
            Platform.ALT_LINUX: (
                'ln -s /etc/httpd2/conf/sites-available/{0} '
                '/etc/httpd2/conf/sites-enabled/{0}'.format(self.cfg_name),),
            Platform.CENTOS: (),
        }
        return copy_commands.get(Platform.distribution)
