# coding: utf-8

from __future__ import unicode_literals

import logging
import subprocess

from tionix_autosetup.executors.base import (
    Executor, ModuleInstallCheckMixin, ModuleUpdateCheckMixin,
    ModuleUninstallCheckMixin)
from tionix_autosetup.helpers import (
    PackageTypesEnum, ManagerTypesEnum, is_tionix_package, LATEST_VERSION,
    TIONIX_LICENSING_VERSION)
from tionix_autosetup.variables import PackageVersion

logger = logging.getLogger(__name__)


class RPMExecutor(Executor):
    packages = None
    command_templates = None
    package_type = PackageTypesEnum.RPM
    duplicate_packages = {
        'python-pip': ('python2-pip', )
    }

    def __init__(self, variable_box=None, **kwargs):
        super(RPMExecutor, self).__init__(variable_box=variable_box, **kwargs)
        self.packages = kwargs.get('packages', self.packages) or []
        self.package_manager = kwargs.get('package_manager',
                                          ManagerTypesEnum.APT)
        self.critical_operation = kwargs.get(
            'critical_operation', self.critical_operation)
        if not isinstance(self.packages, (list, tuple)):
            self.packages = [self.packages]
        self.version = variable_box.PKG_VERSION

    @property
    def commands(self):
        commands_list = [
            self.command_templates.get(self.package_manager) % package
            for package in self.packages if package]
        return commands_list

    @staticmethod
    def _check_return_code(cmd):
        result = False
        try:
            p = subprocess.Popen(cmd, shell=True,
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
            p.communicate()
            result = not bool(p.returncode)
        except OSError:
            pass
        return result

    def pre_check(self, *arg, **kwargs):
        which_cmd = 'which %s' % self.package_manager
        return self._check_return_code(which_cmd)

    def is_package_installed(self, package_name):
        installed = False
        packages = [package_name]
        duplicate_packages = self.duplicate_packages.get(package_name)
        if duplicate_packages:
            packages += duplicate_packages

        for package in packages:
            rpm_cmd = 'dpkg -l {0} || rpm -q {0}'.format(package)
            if self._check_return_code(rpm_cmd):
                installed = True
                break

        return installed


class RPMInstall(ModuleInstallCheckMixin, RPMExecutor):
    critical_operation = True
    valid_codes = 0,

    command_templates = {
        ManagerTypesEnum.APT: 'apt-get -y install %s',
        ManagerTypesEnum.YUM: 'yum -y install %s',
    }

    versions_pointers = {
        ManagerTypesEnum.APT: '=',
        ManagerTypesEnum.YUM: '-',
    }

    @property
    def commands(self):
        commands_list = []
        for package in self.packages:
            if package:
                if is_tionix_package(package):
                    pointer = self.versions_pointers.get(self.package_manager)
                    version = self.get_template_install_version(
                        pointer, package)
                    if version:
                        package += version
                commands_list.append(
                    self.command_templates.get(self.package_manager) % package)

        return commands_list

    def get_template_install_version(self, pointer, package):
        """
        Template for a command to work with specific version of the package.
        """
        version = None
        if 'tionix-licensing' in package:
            version = "%s%s --force-yes" % (pointer, TIONIX_LICENSING_VERSION)
        elif self.version != LATEST_VERSION:
            version = "%s{%s}* --force-yes" % (pointer, PackageVersion.name)
        return version


class RPMUpdate(ModuleUpdateCheckMixin, RPMExecutor):
    critical_operation = True

    command_templates = {
        ManagerTypesEnum.APT: 'apt-get -y install %s',
        ManagerTypesEnum.YUM: 'yum -y update %s',
    }


class RPMUninstall(ModuleUninstallCheckMixin, RPMExecutor):
    critical_operation = True

    command_templates = {
        ManagerTypesEnum.APT: 'apt-get remove -y %s',
        ManagerTypesEnum.YUM: 'yum -y remove %s',
    }
