# coding: utf-8

from __future__ import unicode_literals

import logging
import pkg_resources

from tionix_autosetup.executors.base import (
    Executor, ModuleInstallCheckMixin, ModuleUpdateCheckMixin,
    ModuleUninstallCheckMixin)
from tionix_autosetup.helpers import (
    PackageTypesEnum, increment_version, is_tionix_package, LATEST_VERSION,
    TIONIX_LICENSING_VERSION)
from tionix_autosetup.variables import PackageVersion

logger = logging.getLogger(__name__)


class PIPExecutor(Executor):
    package_type = PackageTypesEnum.PIP
    packages = None
    command_template = None

    update_pip_command = 'pip install -U "pip<10.0"'
    _updated = False

    def __init__(self, variable_box=None, **kwargs):
        super(PIPExecutor, self).__init__(variable_box=variable_box, **kwargs)
        self.packages = kwargs.get('packages', self.packages) or []
        self.critical_operation = kwargs.get(
            'critical_operation', self.critical_operation)
        if not isinstance(self.packages, (list, tuple)):
            self.packages = [self.packages]
        self.version = variable_box.PKG_VERSION

    @property
    def updated(self):
        return PIPInstall._updated

    @updated.setter
    def updated(self, value):
        PIPInstall._updated = value

    @property
    def commands(self):
        commands_list = [
            self.command_template % package
            for package in self.packages if package]
        if not self.updated:
            commands_list.insert(0, self.update_pip_command)
            self.updated = True
        return commands_list

    def pre_check(self, *arg, **kwargs):
        return self.is_package_installed('pip')

    @staticmethod
    def is_package_installed(package_name):
        # seperate package name from pip arguments if present
        package_name = package_name.split(' ')[-1]
        reload(pkg_resources)
        result = True
        try:
            package = pkg_resources.Requirement.parse(package_name)
        except (ValueError, pkg_resources.RequirementParseError):
            result = False

        else:
            package_name = package.project_name
            try:
                pkg_resources.get_distribution(package_name)
            except pkg_resources.DistributionNotFound:
                result = False
            except (pkg_resources.VersionConflict,
                    pkg_resources.ContextualVersionConflict) as err:
                logger.warning('An already-installed version conflicts with '
                               'the requested version: {0}'.format(err))
        return result


class PIPInstall(ModuleInstallCheckMixin, PIPExecutor):
    critical_operation = True
    valid_codes = 0,

    command_template = (
        'export PIP_EXTRA_INDEX_URL=http://{PYPI_HOST}/simple/ && '
        'pip --no-cache-dir install {PIP_ARGS} %s '
        '--trusted-host {PYPI_HOST}')

    def get_template_install_version(self, package):
        """
        Template for a command to work with specific version of the package.
        If version is minor, returns template to install the latest available
        minor version.
        """
        template = ''
        if 'tionix-licensing' in package:
            template = ">={%s}" % TIONIX_LICENSING_VERSION
        elif self.version and self.version != LATEST_VERSION:
            if self.version.count('.') == 4:  # is dev version
                template = "=={%s}"
            else:
                template = "'>={%%s},<%s'" % increment_version(self.version)
            template = template % PackageVersion.name
        return template

    @property
    def commands(self):
        # pylint: disable=duplicate-code
        commands_list = []
        for package in self.packages:
            if package:
                if is_tionix_package(package):
                    package += self.get_template_install_version(package)
                commands_list.append(self.command_template % package)
        if not self.updated:
            commands_list.insert(0, self.update_pip_command)
            self.updated = True

        return commands_list


class PIPUpdate(ModuleUpdateCheckMixin, PIPExecutor):
    critical_operation = True

    command_template = (
        'export PIP_EXTRA_INDEX_URL=http://{PYPI_HOST}/simple/ && '
        'pip install --no-cache-dir -U --no-deps {PIP_ARGS} %s'
        ' --trusted-host {PYPI_HOST}')


class PIPUninstall(ModuleUninstallCheckMixin, PIPExecutor):
    critical_operation = True

    command_template = 'pip uninstall -y %s'
