# coding: utf-8

from __future__ import unicode_literals

import pwd

from tionix_autosetup.executors.base import Executor, PolicyScopeExecutor
from tionix_autosetup.helpers import Platform


class SystemServicesRestart(Executor):

    nova_commands = {
        Platform.UBUNTU: ('service nova-api restart',),
        Platform.ALT_LINUX: ('systemctl restart openstack-nova-api',),
        Platform.CENTOS: ('systemctl restart openstack-nova-api',),
    }

    cinder_commands = {
        Platform.UBUNTU: ('service cinder-api restart',),
        Platform.ALT_LINUX: ('systemctl restart openstack-cinder-api',),
        Platform.CENTOS: ('systemctl restart openstack-cinder-api',),
    }

    heat_commands = {
        Platform.UBUNTU: ('service heat-api restart',),
        Platform.ALT_LINUX: ('systemctl restart openstack-heat-api',),
        Platform.CENTOS: ('systemctl restart openstack-heat-api',),
    }

    supervisor_cmd = 'supervisorctl update && supervisorctl restart all'

    @property
    def commands(self):
        return (
            self.nova_commands.get(Platform.distribution),
            self.cinder_commands.get(Platform.distribution),
            self.heat_commands.get(Platform.distribution),
            self.supervisor_cmd
        )


class MemcachedRestart(Executor):

    memcached_commands = {
        Platform.UBUNTU: ('service memcached restart',),
        Platform.ALT_LINUX: ('systemctl restart memcached',),
        Platform.CENTOS: ('systemctl restart memcached',),
    }

    @property
    def commands(self):
        return self.memcached_commands.get(Platform.distribution)


class SupervisorStart(Executor):

    supervisor_commands = {
        Platform.UBUNTU: ('service supervisor start', 'supervisord',),
        Platform.ALT_LINUX: ('systemctl start supervisord', 'supervisord',),
        Platform.CENTOS: ('systemctl start supervisord', 'supervisord',),
    }

    @property
    def commands(self):
        return (
            'systemctl enable supervisord.service',
            self.supervisor_commands.get(Platform.distribution)
        )


class AddUser(Executor):

    critical_operation = True

    commands = (
        'useradd -m {TNX_USERNAME}',
    )

    @staticmethod
    def _user_exist(username):
        result = True
        try:
            pwd.getpwnam(username)
        except KeyError:
            result = False
        return result

    @property
    def username(self):
        return self.variables_dict.get('TNX_USERNAME')

    def pre_check(self, *arg, **kwargs):
        """ Check if user exists before run

        Check passed if doesn't exist
        """
        return not self._user_exist(self.username)

    def post_check(self, *arg, **kwargs):
        """ Check if user exists after run

        Check passed if user exists
        """
        return self._user_exist(self.username)


class Migration(PolicyScopeExecutor):
    critical_operation = True

    module_name = None
    command_template = 'openstack tnx db migrate -n {module_name}'

    @property
    def commands(self):
        return (
            self.command_template.format(module_name=self.module_name),
        )


class RunConfiguration(PolicyScopeExecutor):
    critical_operation = True

    module_names = None
    command_template = 'openstack tnx configure -n {module_names}'

    @property
    def commands(self):
        names = self.module_names or ''
        if isinstance(names, (tuple, list)):
            names = ' '.join(names)

        return (
            self.command_template.format(module_names=names),
        )
