# coding: utf-8

"""
Helpers methods
"""

import argparse
import collections
import functools
import logging
import os
import re
import shutil
import subprocess
import sys
from StringIO import StringIO
from contextlib import contextmanager
from datetime import datetime

# pylint:disable=no-name-in-module, import-error
from distutils.version import LooseVersion
from platform import linux_distribution

import pkg_resources
from pbr.version import SemanticVersion

from tionix_autosetup import BASE_DIR

logger = logging.getLogger(__name__)
LICENSE_CONTROL_GROUP = u'tionix_license_control'
AUTOSETUP_CONF_DIR = u'/etc/tionix/autosetup/'
BEFORE_AUTOSETUP_ENV = AUTOSETUP_CONF_DIR + u'pip_packages_before_{0}.env'
ACTUAL_AUTOSETUP_ENV = AUTOSETUP_CONF_DIR + u'pip_packages_actual.env'
INSTALL_HISTORY_FILE = AUTOSETUP_CONF_DIR + u'install_history.txt'
LATEST_VERSION = u'last'


class classproperty(property):
    def __get__(self, instance, cls=None):
        getter = getattr(self, 'fget')
        return getter(cls)


def get_package_name(base_dir=None):
    base_dir = base_dir or BASE_DIR
    dirs = reversed(os.path.abspath(base_dir).split('/'))
    try:
        package_name = [dir_name.replace('_', '-') for dir_name in dirs
                        if 'tionix_' in dir_name][0]
    except IndexError:
        package_name = None
    return package_name


def get_version():
    """Get installed package version """
    package_name = get_package_name()
    return get_package_version(package_name)


def brief_version(version, minor=False):
    """Return the short version minus any alpha/beta tags."""
    v = SemanticVersion.from_pip_string(version)
    if minor:
        vs = v.version_tuple()
        result = '{0}.{1}'.format(vs[0], vs[1])
    else:
        result = v.brief_string()
    return result


def increment_version(version):
    """
    Increments last digit of the version.

    :return:
        1.20 => 1.21
        1.21.1 => 1.21.2
        1.21.2.dev4 => 1.21.3
    """
    minor = version.count('.') <= 2
    version = SemanticVersion.from_pip_string(version)
    return version.increment(minor=minor).release_string()


def get_formatted_version():
    return u"{0} {1}".format(
        get_package_name(), get_version()
    )


def create_argparser(*args, **kwargs):
    parser = argparse.ArgumentParser(*args, **kwargs)
    parser.add_argument(
        '--version', action='version', version=get_formatted_version())
    return parser


class OperationsEnum(object):
    INSTALL = u'install'
    UPDATE = u'update'
    UNINSTALL = u'uninstall'
    CHECK_CONNECTION = u'connections_check'
    POST_CHECK = u'post_check'
    ROLLBACK = u'rollback'

    values = (INSTALL, UPDATE, UNINSTALL, CHECK_CONNECTION, POST_CHECK,
              ROLLBACK)


def get_licensed_module_names(group=LICENSE_CONTROL_GROUP, module_names=None):
    if not isinstance(module_names, collections.Iterable):
        module_names = [module_names]
    entry_points = []
    for module_name in module_names:
        entry_points.extend(
            pkg_resources.iter_entry_points(group=group,
                                            name=module_name)
        )
    return [entry_point.name for entry_point in entry_points]


def remove_paths(paths):
    for path in paths:
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path, ignore_errors=True)


def check_module_installed(package_name):
    try:
        pkg_resources.get_distribution(package_name)
    except pkg_resources.DistributionNotFound:
        return False
    except (pkg_resources.VersionConflict,
            pkg_resources.ContextualVersionConflict):
        pass
    return True


class Versions(object):
    MITAKA = '13'
    LIBERTY = '12'


def get_package_version(distr):
    version = None
    try:
        package = pkg_resources.get_distribution(distr)
        version = package.version
    except (TypeError, pkg_resources.DistributionNotFound):
        logger.warning(u'Package not found: {0}'.format(distr))
    except (pkg_resources.VersionConflict,
            pkg_resources.ContextualVersionConflict) as err:
        logger.warning(u'Version conflict found: {0}'.format(err))
    return version


def get_formatting_distr(distr):
    version = get_package_version(distr)
    return version.split('.')[0] if version else ''


def installed_tionix_modules():
    from tionix_autosetup.packages import PACKAGES
    return [module_name for module_name, pip_name in PACKAGES.pip_names.items()
            if check_module_installed(pip_name) and
            module_name != PACKAGES.TIONIX_Autosetup]


@contextmanager
def suppress_stdout():
    new_stdout = StringIO()
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    logging.disable(logging.ERROR)
    try:
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        logging.disable(logging.NOTSET)


class PackageTypesEnum(object):
    PIP = u'pip'
    RPM = u'rpm'
    values = (PIP, RPM)


class ManagerTypesEnum(object):
    PIP = 'pip'
    APT = 'apt-get'
    YUM = 'yum'
    values = (PIP, APT, YUM)


class Platform(object):
    ALT_LINUX = u'alt linux'
    UBUNTU = u'ubuntu'
    CENTOS = u'centos linux'
    DEFAULT = UBUNTU

    DISTRIBUTIONS = (ALT_LINUX, UBUNTU, CENTOS)

    _system_packages_manager = {
        UBUNTU: ManagerTypesEnum.APT,
        ALT_LINUX: ManagerTypesEnum.APT,
        CENTOS: ManagerTypesEnum.YUM,
    }

    _distribution = None

    @classproperty
    def distribution(cls):  # pylint:disable=no-self-argument
        if cls._distribution is None:
            distribution, _version, _id = linux_distribution()

            if distribution is not None:
                cls._distribution = unicode(distribution).lower().strip()

            if cls._distribution not in cls.DISTRIBUTIONS:
                logger.warning(
                    u'Current distribution is not supported: {0}. '
                    u'Distribution is to default: {1}'.format(
                        distribution, cls.DEFAULT))
                cls._distribution = cls.DEFAULT

        return cls._distribution

    @classproperty
    def system_manager(cls):  # pylint:disable=no-self-argument
        return cls._system_packages_manager.get(cls.distribution)


def save_installed_packages_info(file_path=ACTUAL_AUTOSETUP_ENV):
    commands = (
        u'mkdir -p {0}'.format(AUTOSETUP_CONF_DIR),
        u'pip freeze > {0}'.format(file_path),
    )
    for cmd in commands:
        try:
            popen = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
            _stdout, stderr = popen.communicate()
            if stderr:
                logger.warning(u'Error in subprocess: {0}'.format(stderr))
        except OSError as err:
            logger.warning(u'Command failed: {cmd} '
                           u'with error {err}.'.format(cmd=cmd, err=err))

    logger.debug(u'Installed packages info successfully saved '
                 u'to file: {0}'.format(file_path))

    return file_path


def save_installation_source(variable_box, file_path=INSTALL_HISTORY_FILE):
    variable_box = variable_box.to_dict()
    source = variable_box.get('SOURCE')
    host = ''

    if source == ManagerTypesEnum.PIP:
        pip_args = variable_box.get('PIP_ARGS')
        used_args = pip_args.split()
        if '--find-links' in used_args or '-f' in used_args:
            host = '-'
        else:
            host = variable_box.get('PYPI_HOST')
    elif source == ManagerTypesEnum.APT:
        host = ManagerTypesEnum.APT
    if host:
        host_address = '{timestamp:%Y.%m.%d-%H:%M:%S};{host}'.format(
            timestamp=datetime.now(),
            host=host,
        )
        try:
            if not os.path.exists(os.path.dirname(file_path)):
                os.makedirs(file_path)
            with open(file_path, 'a+') as f:
                f.write(host_address + '\n')
        except (IOError, OSError) as err:
            logger.error(u'Unable to save installation source '
                         u'to file: {0}'.format(err))


def fill_missing_keys(original, etalon):
    """
    Compares etalon and original dicts and recursively fills missing keys.

    :param original: dict
    :param etalon: dict
    :return: dict with all keys from etalon
    """
    result_dict = original.copy()

    for key, value in etalon.items():
        if key not in result_dict:
            result_dict[key] = value
        elif isinstance(value, dict):
            result_dict[key] = fill_missing_keys(result_dict[key], value)

    return result_dict


def ask_yes_no_to_continue():
    """Prompts user to answer Yes or No."""
    answer = None
    print "Would you like to continue?"
    while answer not in ('yes', 'y', 'no', 'n'):
        answer = raw_input("Please reply Yes or No:\n").lower()
    return answer in ('yes', 'y')


def get_packages_versions(packages):
    """
    Converts package names in their pip module names and returns
     their versions.
    :returns {module_name: module_version}
    """
    from tionix_autosetup.packages import PACKAGES
    module_names = [PACKAGES.pip_names[pkg] for pkg in packages]
    return {module: get_package_version(module) for module in module_names}


def get_latest_installed_pkg(exclude_pkgs=None):
    """
    Returns the latest version of Tionix package installed in a system.
    Except excluded packages.
    """
    exclude_pkgs = exclude_pkgs or []
    latest_installed_version = ''
    pkgs_installed = set(installed_tionix_modules()) - set(exclude_pkgs)
    if pkgs_installed:
        pkgs_installed_versions = [
            LooseVersion(version) for version
            in get_packages_versions(pkgs_installed).values()
        ]
        latest_installed_version = str(max(pkgs_installed_versions))
    return latest_installed_version


def is_valid_version(version):
    """
    Checks that given string is plausible version number.

    Valid examples:
        0.15
        1.22.1
        1.22.1.dev14
    """
    ver_pattern = r"\d+\.\d+(\.\d+)?(\.\w+\d+)?"
    return bool(re.match(ver_pattern, version))


def is_tionix_package(package, exclude=''):
    """
    Check if package contains 'tionix' in its name
    :param package: pip or rpm package name
    :param exclude: part if pip or rpm package name
    :return: bool
    """
    package = package.lower()
    return 'tionix' in package and exclude not in package


def memoized(func):
    cache = func.cache = {}

    @functools.wraps(func)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    return memoizer


AUTOSETUP_VER = get_version() or ''
AUTOSETUP_BRIEF_VER = brief_version(
    AUTOSETUP_VER, minor=True) if AUTOSETUP_VER else ''
TIONIX_LICENSING_VERSION = '2.0.0'
