# coding: utf-8

from __future__ import unicode_literals

import logging

from tionix_autosetup import variables
from tionix_autosetup.executors.base import AutosetupException
from tionix_autosetup.helpers import OperationsEnum, memoized
from tionix_autosetup.packages import PACKAGES
from tionix_autosetup.parser import get_terminal_width, build_wrapped_table

logger = logging.getLogger(__name__)


@memoized
def get_variables_for_package(package, **kwargs):
    """ Parse available package env variables
    :param package: package name
    :param kwargs: params for VariableBox initialization
    :return: instance of tionix_autosetup.variables:VariableBox
    """
    fabric = PACKAGES.variable_box_fabric[package]
    # WARNING
    # order for update_by_instance reflecting
    # on result values for variables
    variable = fabric(variables.DefaultsVariableBox, **kwargs)
    variable.update_by_instance(fabric(variables.SysEnvVariableBox,
                                       **kwargs))
    if kwargs.get("overwrite_configs"):
        kwargs['config_file'] = PACKAGES.config_path[package]
        variable.update_by_instance(fabric(variables.YamlFileBox,
                                           **kwargs))

    variable.update_by_instance(fabric(variables.ArgsVariableBox,
                                       **kwargs))
    variable.update_by_instance(fabric(variables.FileVariableBox,
                                       **kwargs))

    return variable


def process_pipeline(pipeline, variable_box, package, force=False):
    """ Run all sequence of task in pipeline
    :param pipeline: iterable of callable objects
    :param variable_box: variables
    :param package: processing package name
    :param force: flag to continue if executor processing failed
    :return results: dict with executors results
    :type variable_box: tionix_autosetup.variables.VariableBox
    """
    results = {}
    for executor in pipeline:
        executor_instance = executor(variable_box)
        result = executor_instance.run()
        results[executor] = result

        if executor_instance.critical_operation:
            _results = result
            if not isinstance(result, (list, tuple)):
                _results = [result]
            if any(_result.failed() for _result in _results):
                msg = '{0} package processing aborted.'.format(package)
                if not force:
                    raise AutosetupException(msg)
    return results


def get_pipeline(operation, package):
    if operation == OperationsEnum.INSTALL:
        pipelines = getattr(PACKAGES, 'pipelines', None)
    else:
        attr_name = '{0}_pipelines'.format(operation)
        pipelines = getattr(PACKAGES, attr_name, None)
    if pipelines:
        return pipelines.get(package)


def output_variables_for_package(package, **env_vars):
    variable_box = get_variables_for_package(PACKAGES.TIONIX, **env_vars)
    variables_dict = variable_box.to_dict()
    if package != PACKAGES.TIONIX:
        differences = {}
        variable_box = get_variables_for_package(package, **env_vars)
        for key, value in variable_box.to_dict().items():
            if not key.startswith('MODULE_'):
                if variables_dict.get(key) != value:
                    differences[key] = value
        variables_dict = differences

    if variables_dict:
        headers = [u'N', u'Variable name', u'Value']
        rows = []
        for i, (variable, value) in enumerate(variables_dict.items(), 1):
            if isinstance(value, (list, tuple)):
                value = u", ".join(value)
            rows.append([i, variable, value])

        screen_width = get_terminal_width()
        formatted_table = build_wrapped_table(headers, rows, screen_width)

        table_name = u'{} extracted variables'.format(package)
        table_data = (
            u"{border}\n"
            u"{table_name}\n"
            u"{border}\n"
            u"{table_data}"
        ).format(
            table_name=table_name,
            border="=" * len(table_name),
            table_data=str(formatted_table) + '\n'
        )
        logger.info("\n" + table_data)


def process_package(package, operation, env_vars, force=False):
    logger.info('Processing package {0}. Operation: {1}'.format(
        package, operation))
    variable_box = get_variables_for_package(package, **env_vars)

    if package != PACKAGES.TIONIX:
        output_variables_for_package(package, **env_vars)

    pipeline = get_pipeline(operation, package)
    if pipeline:
        process_pipeline(pipeline, variable_box, package, force=force)
