# coding: utf-8

from __future__ import unicode_literals

import ConfigParser
import glob
import json
import locale
import logging
import os
import re
import shutil

from datetime import datetime

from tionix_autosetup.executors.config import (
    DropConfig, RenameConfig, ConfigWithSectionsParser, UpdateSudoers,
    read_config_file)
from tionix_autosetup.executors.package import Install, Uninstall, Update
from tionix_autosetup.executors.rabbit import DropRabbitUser
from tionix_autosetup.executors.rollback import RollbackExecutor
from tionix_autosetup.executors.base import (
    Executor, AllModulesRemovedMixin, AdminRoleCheckMixin,
    CloudDomainPolicyMixin, AppendPath, Dirs, AutosetupException,
    AutosetupAbortException, PolicyScopeExecutor, AutosetupCommandError)
from tionix_autosetup.executors.service import (
    SystemServicesRestart, SupervisorStart, AddUser,
    MemcachedRestart, RunConfiguration, Migration)
from tionix_autosetup.executors.mysql import (
    DbConnectionCheck, MysqlDropCommands, MysqlCommands)
from tionix_autosetup.executors.rabbit import (
    RabbitMQConnectionCheck, RabbitCommands)
from tionix_autosetup.executors.apache import ApacheRestart
from tionix_autosetup.executors.config import (
    Config, ActualizeConfig, UpdateYamlConfig, DropPath, DropSupervisorConfig)
from tionix_autosetup.helpers import (
    installed_tionix_modules, Platform, PackageTypesEnum,
    OperationsEnum, get_packages_versions, brief_version,
    ask_yes_no_to_continue, get_latest_installed_pkg, is_valid_version,
    AUTOSETUP_BRIEF_VER, LATEST_VERSION)


logger = logging.getLogger(__name__)

yaml_config_name = 'tionix.yaml'
dirs = ['/var/log/tionix/client']
keystone_policy_cfg = 'policy.json'
keystone_dir = '/etc/keystone'
nova_api_paste_conf = '/etc/nova/api-paste.ini'
keystone_cfg = os.path.join(keystone_dir, 'keystone.conf')
rename_date_format = '%d%m%Y_%H%M%S'
DEFAULT_LDAP_DRIVER = 'keystone.identity.backends.ldap.Identity'
TNX_LDAP_DRIVER = 'tionix_client.identity.backends.ldap.TnxIdentity'

PIP_UPGRADE_STRATEGY = '-U --upgrade-strategy only-if-needed %s'

package_names = {
    PackageTypesEnum.PIP: 'tionix-client',
    PackageTypesEnum.RPM: 'python-module-tionix_client',
}

pip_extra_packages = {
    OperationsEnum.INSTALL: [
        'tionix-licensing',
        PIP_UPGRADE_STRATEGY % 'python-subunit',
        PIP_UPGRADE_STRATEGY % 'packaging',
        PIP_UPGRADE_STRATEGY % 'requestsexceptions',
        PIP_UPGRADE_STRATEGY % 'sqlalchemy-migrate',
        PIP_UPGRADE_STRATEGY % 'jsonpath-rw-ext',
    ],
    OperationsEnum.UNINSTALL: [
        'tionix-licensing',
        'tionix-scheduler-client',
    ],
}

rpm_extra_packages = {
    OperationsEnum.INSTALL: [
        'python-module-tionix_licensing',
        'python-module-subunit',
        'python-module-requestsexceptions',
        'python-module-migrate',
        'python-module-jsonpath-rw-ext',
        'python-module-ua_parser',
        'python-module-user_agents',
    ],
    OperationsEnum.UNINSTALL: [
        'python-module-tionix_licensing',
        'python-module-tionix_scheduler_client',
    ],
}

extra_packages = {
    PackageTypesEnum.PIP: pip_extra_packages,
    PackageTypesEnum.RPM: rpm_extra_packages,
}

system_packages = {
    Platform.UBUNTU: (
        'gcc-4.7', 'python-dev', 'python-pip', 'supervisor', 'python-celery',
    ),
    Platform.ALT_LINUX: (
        'gcc4.7', 'python-dev', 'python-module-pip', 'supervisor',
        'python-module-celery', 'python-module-django-dbbackend-mysql',
    ),
    Platform.CENTOS: (
        'gcc', 'python-devel', 'python-pip', 'supervisor',
    ),
}


horizon_paths = {
    Platform.UBUNTU: '/usr/share/openstack-dashboard/'
                     'openstack_dashboard/conf',
    Platform.ALT_LINUX: '/etc/openstack-dashboard',
    Platform.CENTOS: '/etc/openstack-dashboard',
}


class BaseConfig(Config):
    config_path = '/etc/tionix/tionix.yaml'
    template_name = 'base_config.template'


class ClientDirs(Dirs):
    dirs = dirs


class ClientDropLogFiles(DropPath):
    paths = dirs


class ConfigJournalNovaWorker(ConfigWithSectionsParser):
    config_path = '/etc/supervisor/conf.d/tnx_nova_journal.conf'
    config_section = 'program:tionix-nova-journal'
    config_keys = {
        'user': '{TNX_USERNAME}',
        'group': '{TNX_USERNAME}',
    }


class ConfigJournalKeystoneWorker(ConfigWithSectionsParser):
    config_path = '/etc/supervisor/conf.d/tnx_keystone_journal.conf'
    config_section = 'program:tionix-keystone-journal'
    config_keys = {
        'user': '{TNX_USERNAME}',
        'group': '{TNX_USERNAME}',
    }


class UpdateBaseYamlConfig(UpdateYamlConfig, BaseConfig):
    module_config = False


class TionixActualizeConfig(ActualizeConfig):
    config_name = yaml_config_name


class AssignAdminRole(CloudDomainPolicyMixin, AdminRoleCheckMixin):
    critical_operation = True

    @property
    def commands(self):
        return ('openstack role add --user {DOMAIN_USER} '
                '--domain {ADMIN_DOMAIN_ID} admin',)

    def pre_check(self, *arg, **kwargs):
        result = False
        if self.cloud_domain_policy:
            try:
                exist = self.check_domain_admin_role()
            except AutosetupException as err:
                logger.error(err)
            else:
                if not exist:
                    result = True
        return result

    def post_check(self, *arg, **kwargs):
        result = True
        if self.cloud_domain_policy:
            try:
                exist = self.check_domain_admin_role()
            except AutosetupException as err:
                logger.error(err)
            else:
                if not exist:
                    logger.critical(u'Admin role is not assigned to user')
                    result = False
        return result


class BackupKeystonePolicyConfig(CloudDomainPolicyMixin, Executor):
    critical_operation = True
    config_name = keystone_policy_cfg
    config_dir = keystone_dir
    renaming_postfix = '.' + datetime.now().strftime(rename_date_format)
    _keystone_cfg_path = os.path.join(keystone_dir, keystone_policy_cfg)
    _keystone_cfg_path_renamed = _keystone_cfg_path + renaming_postfix

    commands = (
        'cp {0} {1}'.format(_keystone_cfg_path, _keystone_cfg_path_renamed),
    )


class RestoreKeystonePolicyConfig(
        AllModulesRemovedMixin, CloudDomainPolicyMixin, Executor):
    critical_operation = True
    config_name = keystone_policy_cfg
    config_dir = keystone_dir
    _keystone_cfg_path = os.path.join(keystone_dir, keystone_policy_cfg)

    def _get_noncloud_backup(self):
        """Get valid backup without cloud admin policy created by Autosetup"""

        pattern = os.path.join(self.config_dir, self.config_name + '.*')
        parsed_backups = {}

        for backup_file in glob.glob(pattern):
            try:
                if not self._is_cloud_config(backup_file):
                    date_str = backup_file.split(self.config_name + '.')[-1]
                    parsed_backups[backup_file] = datetime.strptime(
                        date_str, rename_date_format)
            except (OSError, ValueError):
                continue

        result = sorted(parsed_backups, key=parsed_backups.get)

        return unicode(result[-1]) if result else None

    def post_check(self, *arg, **kwargs):
        return not self._is_cloud_config(self._keystone_cfg_path)

    @staticmethod
    def _is_cloud_config(config_name):
        with open(config_name) as fp:
            data = json.load(fp)

        return "cloud_admin" in data

    def _run(self, **kwargs):
        found_backup = self._get_noncloud_backup()
        if found_backup:
            shutil.copy(found_backup, self._keystone_cfg_path)


class CreateKeystonePolicyConfig(CloudDomainPolicyMixin, Config):
    critical_operation = True
    config_path = os.path.join(keystone_dir, keystone_policy_cfg)
    template_name = 'policy_v3_cloud_sample.template'
    rewrite = True


class CopyKeystonePolicyConfig(CloudDomainPolicyMixin, Executor):
    critical_operation = True
    _keystone_policy_cfg_horizon = 'keystone_policy.v3cloudsample.json'

    _keystone_cfg_path = os.path.join(keystone_dir, keystone_policy_cfg)

    horizon_path = horizon_paths.get(Platform.distribution)

    def pre_check(self, *arg, **kwargs):
        check = super(CopyKeystonePolicyConfig, self).pre_check(*arg, **kwargs)
        if check and not os.path.exists(self.horizon_path):
            logger.info(
                'Horizon configuration file path not found: %s',
                self.horizon_path)
            check = False
        return check

    def _run(self, **kwargs):
        keystone_horizon_cfg_path = os.path.join(
            self.horizon_path, self._keystone_policy_cfg_horizon)

        if os.path.exists(keystone_horizon_cfg_path):
            os.remove(keystone_horizon_cfg_path)
        shutil.copyfile(self._keystone_cfg_path, keystone_horizon_cfg_path)
        stat_info = os.stat(os.path.dirname(keystone_horizon_cfg_path))
        os.chown(keystone_horizon_cfg_path, stat_info.st_uid, stat_info.st_gid)
        file_permissions = 0644  # rw-r--r--
        os.chmod(keystone_horizon_cfg_path, file_permissions)


class ActivateLdapDriver(ConfigWithSectionsParser):

    use_ldap = False
    config_path = os.path.join(keystone_dir, 'keystone.conf')
    config_section = 'identity'
    domain_config_key = 'domain_config_dir'
    driver_config_key = 'driver'
    _domain_config_dir = None

    original_driver = DEFAULT_LDAP_DRIVER
    updated_driver = TNX_LDAP_DRIVER

    config_keys = {
        driver_config_key: TNX_LDAP_DRIVER,
    }

    def __init__(self, variable_box=None, **kwargs):
        super(ActivateLdapDriver, self).__init__(variable_box, **kwargs)
        self.ldap_domains = []
        if hasattr(variable_box, 'kwargs'):
            self.use_ldap = variable_box.kwargs.get('use_ldap', False)

    def domain_configs(self):
        for r, _d, f in os.walk(self.domain_config_dir):
            for fname in f:
                if re.match(r'keystone\.[^.]+\.conf', fname):
                    config_path = os.path.join(r, fname)
                    yield config_path

    @property
    def domain_config_dir(self):
        if self._domain_config_dir is None:
            read_config_file(self.config_path, config_parser=self.config)
            try:
                self._domain_config_dir = self.config.get(
                    self.config_section, self.domain_config_key)
            except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
                pass
        return self._domain_config_dir

    def _run(self, **kwargs):
        for config_path in self.domain_configs():
            self.config = read_config_file(config_path)
            try:
                conf_val = self.config.get(self.config_section,
                                           self.driver_config_key)
            except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
                pass
            else:
                if conf_val == self.original_driver:
                    self.ldap_domains.append(config_path)
                    self.config_path = config_path
                    super(ActivateLdapDriver, self)._run(**kwargs)

    def pre_check(self, *arg, **kwargs):
        if not self.use_ldap:
            return False

        result = True
        if not os.path.exists(self.config_path):
            logger.error("File {0} doesn't exist".format(self.config_path))
            result = False
        elif not self.domain_config_dir:
            logger.error("Domain config directory doesn't exist in file {0}"
                         "".format(self.config_path))
            result = False
        elif not os.path.exists(self.domain_config_dir):
            logger.error("Directory {0} doesn't exist"
                         "".format(self.domain_config_dir))
            result = False
        return result

    def post_check(self, *arg, **kwargs):
        for config_path in self.ldap_domains:
            config = read_config_file(config_path)
            driver = config.get(self.config_section, self.driver_config_key)
            if driver != self.updated_driver:
                return False
        return True


class DeactivateLdapDriver(AllModulesRemovedMixin, ActivateLdapDriver):
    original_driver = TNX_LDAP_DRIVER
    updated_driver = DEFAULT_LDAP_DRIVER

    config_keys = {
        ActivateLdapDriver.driver_config_key: DEFAULT_LDAP_DRIVER,
    }


class UpdateLocaleVariable(Executor):

    def _run(self, **kwargs):
        lc_all = os.getenv('LC_ALL')
        if not lc_all:
            lang_code, encoding = locale.getdefaultlocale()
            lang = '{0}.{1}'.format(lang_code, encoding)
            os.environ['LC_ALL'] = os.getenv('LANG') or lang


class UpdateNovaVersionConf(ConfigWithSectionsParser):
    config_path = '/etc/nova/nova.conf'
    config_section = 'osapi_v3'
    config_keys = {
        'enabled': 'True',
    }


class EnableNovaNotifications(ConfigWithSectionsParser):
    config_path = '/etc/nova/nova.conf'
    config_section = 'DEFAULT'
    config_keys = {
        'notification_driver': 'messagingv2',
        'notification_topics': 'notifications'
    }


class EnableKeystoneCADF(ConfigWithSectionsParser):
    config_path = keystone_cfg
    config_section = 'DEFAULT'
    config_keys = {
        'notification_format': 'cadf',
    }


class SetKeystoneNotificationsDriver(ConfigWithSectionsParser):
    config_path = keystone_cfg
    config_section = 'oslo_messaging_notifications'
    config_keys = {
        'driver': 'messagingv2',
    }


class BackupNovaConfig(Executor):
    critical_operation = True
    renaming_postfix = '.' + datetime.now().strftime(rename_date_format)
    config_path_renamed = nova_api_paste_conf + renaming_postfix

    commands = (
        'cp {0} {1}'.format(nova_api_paste_conf, config_path_renamed),
    )


class BackupKeystoneConfig(Executor):
    critical_operation = True
    renaming_postfix = '.' + datetime.now().strftime(rename_date_format)
    config_path_renamed = keystone_cfg + renaming_postfix

    commands = (
        'cp {0} {1}'.format(keystone_cfg, config_path_renamed),
    )


class AddAuditFilter(Executor):
    config_path = nova_api_paste_conf
    filter_name = 'tnx_audit'
    filter_section = 'filter:{0}'.format(filter_name)
    filter_key = 'paste.filter_factory'
    filter_val = 'tionix_client.journal.api_filter:filter_factory'
    api_section = 'composite:openstack_compute_api_v21'
    api_key = 'keystone'

    def read_config(self):
        conf = ConfigParser.ConfigParser()
        if not conf.read(self.config_path):
            raise AutosetupCommandError(
                'Failed to read config file: {0}'.format(self.config_path))
        return conf

    def get_api_filters(self, conf):
        try:
            api_pipeline = conf.get(self.api_section, self.api_key)
        except ConfigParser.Error as err:
            raise AutosetupCommandError(
                'Failed to configure audit middleware: {0}'.format(err))
        api_filters = api_pipeline.split()
        return api_filters

    def save_conf(self, conf):
        # noinspection PyTypeChecker
        with open(self.config_path, 'w') as fh:
            conf.write(fh)

    def _run(self, **kwargs):
        conf = self.read_config()
        if not conf.has_section(self.filter_section):
            conf.add_section(self.filter_section)

        conf.set(self.filter_section, self.filter_key, self.filter_val)

        api_filters = self.get_api_filters(conf)
        if len(api_filters) > 1 and self.filter_name not in api_filters:
            api_filters.insert(len(api_filters) - 1, self.filter_name)
        conf.set(self.api_section, self.api_key, ' '.join(api_filters))

        self.save_conf(conf)


class RemoveAuditFilter(AllModulesRemovedMixin, AddAuditFilter):
    def _run(self, **kwargs):
        conf = self.read_config()
        api_filters = self.get_api_filters(conf)
        if self.filter_name in api_filters:
            api_filters.remove(self.filter_name)

        conf.set(self.api_section, self.api_key, ' '.join(api_filters))

        if conf.has_section(self.filter_section):
            conf.remove_section(self.filter_section)

        self.save_conf(conf)


class TionixDirs(Dirs):
    dirs = [
        '/etc/tionix/',
        '/var/log/tionix/',
    ]


class CheckInstallVersionCompatibility(Executor):
    """
    Checks that install version is correct and compatible with other packages.
    """
    critical_operation = True

    def __init__(self, variable_box=None, **kwargs):
        super(CheckInstallVersionCompatibility, self).__init__(variable_box,
                                                               **kwargs)
        self.version = variable_box.PKG_VERSION

    def pre_check(self, *arg, **kwargs):
        return self.version is not None and self.version != LATEST_VERSION

    def _run(self, **kwargs):
        if not is_valid_version(self.version):
            raise ValueError('Invalid package '
                             'version "{0}".'.format(self.version))

        if not self.check_compatibility():
            raise AutosetupAbortException("Compatibility check failed.")

        return

    def check_compatibility(self):
        """
        Checks whether TIONIX.Autosetup and other packages are the same
        version as the one that user is trying to install. If not, warns user
        about possible problems.
        """
        from tionix_autosetup.packages import PACKAGES
        result = True

        pkgs = self.extra_kwargs.get('packages')
        incompatible_pkg = self.extra_kwargs.get('incompatible_pkg', False)
        minor_install_pkg_ver = brief_version(self.version, minor=True)
        if minor_install_pkg_ver != AUTOSETUP_BRIEF_VER:
            logger.warning("TIONIX.Autosetup version ({0}) differs from the "
                           "package version you're trying to install ({1}). "
                           "It's recommended to use TIONIX.Autosetup of the "
                           "same version.".format(AUTOSETUP_BRIEF_VER,
                                                  self.version))
            if not incompatible_pkg:
                result = ask_yes_no_to_continue()

        if result:
            pkg_ltst_inst_version = get_latest_installed_pkg(
                exclude_pkgs=pkgs + [PACKAGES.TIONIX_Autosetup])

            if (pkg_ltst_inst_version and
                    minor_install_pkg_ver !=
                    brief_version(pkg_ltst_inst_version, minor=True)):
                logger.warning("Other package's version ({0}) differs from "
                               "the one you're trying to install. It's "
                               "recommended to install version {1}."
                               "".format(self.version, pkg_ltst_inst_version))
                if not incompatible_pkg:
                    result = ask_yes_no_to_continue()

        return result


class CheckModulesCompatibility(Executor):
    """Check that all modules has the same version as installed version."""
    critical_operation = False

    def __init__(self, variable_box=None, **kwargs):
        super(CheckModulesCompatibility, self).__init__(variable_box, **kwargs)
        self.version = variable_box.PKG_VERSION

    @staticmethod
    def check_versions_compatibility(packages, compare_version):
        check_result = True

        pkg_versions = get_packages_versions(packages)
        brief_compare_version = brief_version(compare_version, minor=True)
        incomp_pkgs = [
            "%s: %s" % (pkg_name, ver)
            for pkg_name, ver in pkg_versions.items()
            if brief_version(ver, minor=True) != brief_compare_version
        ]
        if incomp_pkgs:
            logger.warning(
                "The following packages might have problems with the "
                "installed version ({0}): {1}.".format(compare_version,
                                                       ", ".join(incomp_pkgs)))
            check_result = False

        return check_result

    def _run(self, **kwargs):
        result = True
        installed_packages = installed_tionix_modules()
        if installed_packages and self.version:
            result = self.check_versions_compatibility(installed_packages,
                                                       self.version)
        return result


class RunSelfDiagnosticTest(PolicyScopeExecutor):
    result_log_level = logging.INFO

    commands = (
        'openstack tnx tests',
    )

    def pre_check(self, *arg, **kwargs):
        return installed_tionix_modules()


class TionixYamlRename(RenameConfig):
    config_name = yaml_config_name

    @property
    def paths(self):
        config_paths = []
        if not installed_tionix_modules():
            config_paths = super(TionixYamlRename, self).paths
        return config_paths


class TionixYamlDrop(DropConfig):
    config_name = yaml_config_name

    @property
    def paths(self):
        config_paths = []
        if not installed_tionix_modules():
            config_paths = super(TionixYamlDrop, self).paths
        return config_paths


class InstallTionixClient(Install):
    package_names = package_names


class UpdateTionixClient(Update):
    package_names = package_names


class UninstallTionixClient(AllModulesRemovedMixin, Uninstall):
    package_names = package_names


class InstallExtraTionix(Install):
    critical_operation = False
    package_names = extra_packages


class UninstallExtraTionix(AllModulesRemovedMixin, Uninstall):
    package_names = extra_packages


class CommonSystemPackages(Install):
    default_manager = Platform.system_manager
    packages = system_packages.get(Platform.distribution)


class ClientConfiguration(RunConfiguration):
    module_names = 'tnx_client'


class ClientMigration(Migration):
    module_name = 'tnx_client'


class AptGetUpdate(Executor):
    commands = (
        'which apt-get && apt-get update',
    )


class ClientRabbitCommands(RabbitCommands):
    @property
    def commands(self):
        vhosts = {
            self.variables_dict.get('NOVA_RABBIT_VHOST'),
            self.variables_dict.get('KEYSTONE_RABBIT_VHOST'),
        }
        commands = [
            'rabbitmqctl add_user {RABBIT_USERNAME} {RABBIT_PASSWORD}',
            'rabbitmqctl set_permissions {RABBIT_USERNAME} ".*" ".*" ".*"',
        ]
        for vhost in vhosts:
            commands.append(
                'rabbitmqctl set_permissions -p %s {RABBIT_USERNAME} '
                '".*" ".*" ".*"' % vhost
            )
        return commands


class DropNovaJournalSupervisorConfig(DropSupervisorConfig):
    module_config_prefix = 'tnx_nova_journal'


class DropKeystoneJournalSupervisorConfig(DropSupervisorConfig):
    module_config_prefix = 'tnx_keystone_journal'


pipeline = [
    AppendPath,
    CheckInstallVersionCompatibility,
    AptGetUpdate,
    CommonSystemPackages,
    AddUser,
    UpdateSudoers,
    TionixDirs,
    UpdateLocaleVariable,
    InstallTionixClient,
    InstallExtraTionix,
    ClientDirs,
    ClientConfiguration,
    ConfigJournalNovaWorker,
    ConfigJournalKeystoneWorker,
    BaseConfig,
    UpdateBaseYamlConfig,
    TionixActualizeConfig,
    MysqlCommands,
    ClientRabbitCommands,
    ClientMigration,
    AssignAdminRole,
    BackupNovaConfig,
    UpdateNovaVersionConf,
    EnableNovaNotifications,
    ActivateLdapDriver,
    AddAuditFilter,
    BackupKeystoneConfig,
    EnableKeystoneCADF,
    SetKeystoneNotificationsDriver,
    BackupKeystonePolicyConfig,
    CreateKeystonePolicyConfig,
    CopyKeystonePolicyConfig,
    SupervisorStart,
    ApacheRestart,
    SystemServicesRestart,
]

update_pipeline = [
    UpdateTionixClient,
    ClientDirs,
    ClientConfiguration,
    ClientMigration,
    UpdateBaseYamlConfig,
    TionixActualizeConfig,
    AppendPath,
    AssignAdminRole,
    BackupKeystonePolicyConfig,
    CreateKeystonePolicyConfig,
    CopyKeystonePolicyConfig,
    BackupNovaConfig,
    UpdateNovaVersionConf,
    EnableNovaNotifications,
    AddAuditFilter,
    BackupKeystoneConfig,
    ActivateLdapDriver,
    EnableKeystoneCADF,
    SetKeystoneNotificationsDriver,
    ApacheRestart,
    SystemServicesRestart,
]

connections_check_pipeline = [
    AppendPath,
    DbConnectionCheck,
    RabbitMQConnectionCheck,
]

post_check_pipeline = [
    CheckModulesCompatibility,
    RunSelfDiagnosticTest,
]

rollback_pipeline = [
    RollbackExecutor,
    ApacheRestart,
    MemcachedRestart,
    SystemServicesRestart,
]

uninstall_pipeline = [
    UninstallTionixClient,
    UninstallExtraTionix,
    ClientDropLogFiles,
    TionixYamlRename,
    TionixYamlDrop,
    AppendPath,
    DropRabbitUser,
    BackupKeystonePolicyConfig,
    RestoreKeystonePolicyConfig,
    DeactivateLdapDriver,
    MysqlDropCommands,
    BackupNovaConfig,
    RemoveAuditFilter,
    DropKeystoneJournalSupervisorConfig,
    DropNovaJournalSupervisorConfig,
    ApacheRestart,
    MemcachedRestart,
    SystemServicesRestart,
]
